#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright 2013 Jens Lechtenbörger
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Classes to analyze Scapy traceroutes.

See README.txt.
"""
import csv
import logging
import os
import socket
import struct
import sys


METHODS = set()
try:
    import pygeoip
    METHODS.add("pygeoip")
except ImportError:
    print "Module pygeoip not found.  Not usable."

try:
    import TorCtl
    METHODS.add("torctl")
except ImportError:
    print "Module TorCtl not found.  Not usable."

NETCAT = "/usr/local/bin/netcat"
CYMRUSERVER = "whois.cymru.com"
CYMRUPORT = 43
METHODS.add("cymru")
if not os.path.isfile(NETCAT):
    print """GNU netcat not found.  Thus, whois lookups cannot be performed.
This will only work if AS information from a previous lookup is available.
"""


class IPAddress(object):
    """An IP address with additional information."""

    country = "DE"

    def __init__(self, ipstring):
        """Create IP address from dotted string."""
        self.ipstring = ipstring
        self.ipint = struct.unpack("!L", socket.inet_aton(ipstring))[0]
        self.prefix = None
        self.asno = None
        self.asname = None
        self.registry = None
        self.country = None
        self.country_details = dict()
        self.isforeign = False
        self.isboomeranged = False
        self.torname = None
        self.ixps = set()
        self.route = list()

    def __eq__(self, other):
        """Check equality of self with other."""
        return self.ipstring == other.ipstring

    def __hash__(self):
        """Hash self's string."""
        return hash(self.ipstring)

    def __str__(self):
        if self.torname is not None:
            return "{0} ({1})".format(self.torname, self.ipstring)
        return self.ipstring

    def tostring(self):
        """Return verbose string for self."""
        result = str(self)
        if self.country is not None and self.country.find(" ") > 0:
            result += ";{0}".format(str(self.country_details))
        else:
            result += ";{0}".format(self.country)
        if len(self.ixps) > 0:
            ixpstring = ",".join([ixp.name for ixp in self.ixps])
            result += " (via IXP {0})".format(ixpstring)
        return result

    def routestr(self):
        """Return self's route as string."""
        route = " → ".join([hop.tostring() for hop in self.route])
        return "{0}: {1}".format(self, route)

    def add_country(self, country, method):
        """Register country obtained via method for self."""
        country = country.upper()
        self.country_details[method] = country
        if self.country is None:
            self.country = country
        elif self.country != country:
            logging.warning(
                "Inconsistent country for %s: %s",
                self, str(self.country_details))
            self.country += " {0}".format(country)
        if self.country is not None and self.country != IPAddress.country:
            self.isforeign = True

    def add_ixp(self, ixp):
        """Register that ixp is passed from "here"."""
        self.ixps.add(ixp)

    def passes_ixp(self):
        """Return True if self is reached via IXP from "here"."""
        return len(self.ixps) > 0

    def append_hop(self, hopipaddr):
        """Append hopipaddr to self's route."""
        self.route.append(hopipaddr)
        if hopipaddr.isforeign:
            self.isforeign = True
        elif len(self.route) > 1:
            previous = self.route[-2]
            if (previous.country is not None
                and previous.country != IPAddress.country):
                logging.warning(
                    "Boomerang returning for %s at %s.", self, hopipaddr)
                self.isboomeranged = True

    @staticmethod
    def aton(ipstring):
        """Convert ipstring in dotted string notation to integer."""
        return IPAddress(ipstring).ipint

    @staticmethod
    def ntoa(ipint):
        """Convert IP address ipint from integer to dotted string."""
        return socket.inet_ntoa(struct.pack("!L", ipint))

    @staticmethod
    def from_whois(line):
        """Create IPAddress from a line in Team Cymru whois result."""
        fields = [field.strip() for field in line.split("|")]
        if len(fields) != 7:
            logging.error("Unexpected number of fields in line: %s", line)
            return None
        result = IPAddress(fields[1])
        if fields[0] != "NA":
            result.asno = fields[0]
            result.asname = fields[6]
        if fields[2] != "NA":
            result.prefix = IPPrefix(fields[2])
        if len(fields[3]) > 0:
            result.add_country(fields[3], "cymru")
        if len(fields[4]) > 0:
            result.registry = fields[4]
        return result


class IPPrefix(object):
    """Represent an IP prefix such as 192.168.178.0/23."""

    def __init__(self, prefixstring):
        self.prefixstring = prefixstring
        self.valid = False
        if prefixstring.find("/") > 0:
            ipaddr, length = prefixstring.split("/")
            self.valid = True
        else:
            ipaddr = "0.0.0.1"
            length = "32"
            if prefixstring != "NA":
                logging.warning("Invalid prefix %s.", prefixstring)
        self.ipstart = IPAddress.aton(ipaddr)
        self.length = int(length)
        self.mask = ((1 << self.length) - 1) << (32 - self.length)

    def contains(self, ipstring):
        """Check whether self contains ipstring."""
        if not self.valid:
            return False
        ipcand = IPAddress.aton(ipstring) & self.mask
        return ipcand == self.ipstart

    def __eq__(self, other):
        """Check equality of self with other."""
        return self.ipstart == other.ipstart and self.length == other.length

    def __str__(self):
        """Standard string representation."""
        return self.prefixstring

    def __hash__(self):
        """Hash self's string."""
        return hash(self.prefixstring)


def scapy_clean(trace):
    """Remove shared hops from Scapy trace."""
    hopdict = dict()
    for (_, route) in trace.iteritems():
        for (hopno, (hopipstring, _)) in route.iteritems():
            if hopno not in hopdict:
                hopdict[hopno] = set()
            hopdict[hopno].add(hopipstring)
    hops = hopdict.keys()
    hops.sort()
    unique = 0
    for hop in hops:
        if len(hopdict[hop]) == 1:
            unique = hop
        else:
            break
    if unique > 0:
        msg = """Deleting the first {0} hops from traceroute data as they are
common to all routes.  You may want to verify manually that none of
those appears to be dangerous.""".format(unique)
        logging.warning(msg)
        for (_, route) in trace.iteritems():
            for hopno in range(1, unique + 1):
                if hopno in route:
                    logging.debug(
                        "Deleting shared hop %s with IP address: %s",
                        hopno, route[hopno])
                    del route[hopno]


def scapy_ipstrings(trace, remove_common=False):
    """Return set of ipstrings appearing in Scapy trace.

    If remove_common is True, then initial hops that are common to all traces
    are removed.  This may be dangerous.
    """
    result = set()
    if remove_common:
        scapy_clean(trace)
    for (ipstring, route) in trace.iteritems():
        result.add(ipstring)
        for (_, (hopipstring, _)) in route.iteritems():
            result.add(hopipstring)
    return result


class CymruLookup(object):
    """Lookup information for IP addresses via Team Cymru's whois service."""

    DB = "asinfo.txt"
    WHOIS_OUT = "whois-result.txt"
    WHOIS_IN = "whois-input.txt"

    def __init__(self, datadir="data"):
        self.dbfile = os.path.join(datadir, CymruLookup.DB)
        self.whois_out = os.path.join(datadir, CymruLookup.WHOIS_OUT)
        self.whois_in = os.path.join(datadir, CymruLookup.WHOIS_IN)
        self.ipinfo = dict()
        self.asinfo = dict()
        if os.path.isfile(self.dbfile):
            self._parse(self.dbfile)

    def _parse(self, filename, append=False):
        """Parse whois result in filename and update self.ipinfo.

        If append is True, append the whois result to self.dbfile.
        """
        logging.debug("Parsing whois file %s.", filename)
        with open(filename) as asfile:
            lines = asfile.readlines()
            # First line is a comment.
            if append and not os.path.isfile(self.dbfile):
                with open(self.dbfile, "w") as dbfile:
                    dbfile.write("Team Cymru AS Information\n")

            with open(self.dbfile, "a") as dbfile:
                for line in lines[1:]:
                    if append:
                        dbfile.write(line)
                    ipaddr = IPAddress.from_whois(line)
                    if ipaddr is None:
                        continue
                    self.ipinfo[ipaddr.ipstring] = ipaddr
                    if (ipaddr.asno is not None
                        and ipaddr.asno not in self.asinfo):
                        self.asinfo[ipaddr.asno] = ipaddr.asno
        logging.debug("Parsing done.")

    def _write(self, ipstrings):
        """Write ipstrings to file, suitable for Team Cymru whois lookup."""
        logging.debug("Writing IP addresses.")
        with open(self.whois_in, "w") as whoisfile:
            whoisfile.write("""begin
verbose
noheader
""")
            for ipstring in ipstrings:
                whoisfile.write("{0}\n".format(ipstring))
            whoisfile.write("end\n")

    def whois(self, netcat=NETCAT, server=CYMRUSERVER, port=CYMRUPORT):
        """Invoke whois via netcat to server on port."""
        cmd = "{0} {1} {2} < {3} > {4}".format(
            netcat, server, port, self.whois_in, self.whois_out)
        logging.debug("Invoking whois: %s", cmd)
        os.system(cmd)
        logging.debug("Whois done.")

    def lookup(self, ipstrings):
        """Lookup whois information for ipstrings.

        The result is a dictionary mapping each of the ipstrings to its
        IPAddress instance, hopfully, with information about IP prefix, AS,
        and country.
        """
        unknown = [ipstring for ipstring in ipstrings
                   if ipstring not in self.ipinfo]
        if len(unknown) > 0:
            self._write(unknown)
            self.whois()
            self._parse(self.whois_out, append=True)
        else:
            logging.info("No unknown IP addresses.  Not invoking whois.")
        return dict(filter(
            lambda arg: arg[0] in ipstrings, self.ipinfo.iteritems()))


class IPLookup(object):
    """Lookup for IP addresses."""

    def __init__(self, methods):
        self.methods = set(methods.keys())
        if not self.methods <= METHODS:
            print "\nUnsupported method for IPLookup."
            print "Requested {0} is no subset of supported {1}.".format(
                self.methods, METHODS)
            sys.exit(1)
        self._lookup_functions = dict()
        self._init_functions = {
            "pygeoip": IPLookup._init_pygeoip,
            "torctl": IPLookup._init_torctl,
            "cymru": IPLookup._init_cymru}
        self._geoip = None
        self._tor = None
        self._cymru = None
        for (method, arg) in methods.iteritems():
            self._init_functions[method](self, arg)

    def _init_pygeoip(self, geoipfile):
        """Initialize pygeoip."""
        self._lookup_functions["pygeoip"] = self._lookup_geoip
        if not os.path.isfile(geoipfile):
            logging.error(
                "GeoIP file %s not found.  Not using pygeoip.", geoipfile)
            self.methods.remove("pygeoip")
        else:
            self._geoip = pygeoip.GeoIP(geoipfile, pygeoip.MEMORY_CACHE)

    def _init_torctl(self, passphrase=None):
        """Initialize TorCtl."""
        self._lookup_functions["torctl"] = self._lookup_torctl
        self._tor = TorCtl.connect(passphrase=passphrase)
        if self._tor is None:
            logging.error("Could not initialize Tor control.  Not using Tor.")
            self.methods.remove("torctl")

    def _init_cymru(self, *args):
        """Initialize CymruLookup."""
        self._cymru = CymruLookup()
        # Following unused.  See TODO in lookup().
        #self._lookup_functions["cymru"] = self._lookup_cymru

    def lookup(self, ipstrings, **kwargs):
        """Lookup information for ipstrings.

        If "methods" is contained in kwargs, it specifies the lookup methods
        to use; otherwise, all known methods are used.

        The result is a dictionary mapping each of the given ipstrings to
        an IPAddress instance.
        """
        if "methods" not in kwargs:
            methods = self.methods.copy()
        else:
            methods = kwargs["methods"]
        if "cymru" in methods:
            # TODO Generalize hard-coded bulk lookup?
            # Currently, cymru is the only method with bulk support.
            result = self._lookup_cymru(ipstrings)
            methods.remove("cymru")
        else:
            result = dict()
        for ipstring in ipstrings:
            if ipstring in result:
                ipaddr = result[ipstring]
            else:
                ipaddr = IPAddress(ipstring)
                result[ipstring] = ipaddr
            for method in methods:
                country = self._lookup_functions[method](ipstring)
                if country is None:
                    logging.info(
                        "No country for %s via %s.", ipstring, method)
                else:
                    logging.debug("%s: %s in %s", method, ipstring, country)
                    ipaddr.add_country(country, method)
        return result

    def _lookup_torctl(self, ipstring):
        """Lookup country for ipstring via TorCtl."""
        query = "ip-to-country/{0}".format(ipstring)
        answer = self._tor.get_info(query)
        if answer is not None and query in answer:
            if answer[query] != "??":
                return answer[query]
        return None

    def _lookup_geoip(self, ipstring):
        """Lookup country for ipstring via pygeoip."""
        result = self._geoip.country_code_by_addr(ipstring)
        if result is not None and len(result) > 0:
            return result
        return None

    def _lookup_cymru(self, ipstrings):
        """Lookup ipstrings via Team Cymru's whois."""
        return self._cymru.lookup(ipstrings)


class IXP(object):
    """Representation of an Internet eXchange Point."""

    def __init__(self, ixpno, name):
        self.ixpno = ixpno
        self.name = name
        self.prefixes = set()

    def __str__(self):
        return "{0} ({1})".format(self.ixpno, self.name)

    def add(self, prefixstring):
        """Record information of prefixstring."""
        self.prefixes.add(IPPrefix(prefixstring))


class IXPSearch(object):
    """Search for Internet eXchange Points."""

    def __init__(self, ixpfile="ixp-list.txt", ixpprefixes="ixp-prefixes.txt",
                 **kwargs):
        if (not os.path.isfile(ixpfile)
            or not os.path.isfile(ixpprefixes)):
            print """At least one file with IXP information is missing.
Use correct paths or download: http://www-rp.lip6.fr/~augustin/ixp/"""
            sys.exit(1)
        self.iplookup = IPLookup(kwargs)

        self.ixps = dict()
        with open(ixpfile) as ixpf:
            for ixp in ixpf.readlines():
                ixpno, name = ixp.split("\t")[:2]
                self.ixps[ixpno] = IXP(ixpno, name)
        with open(ixpprefixes) as ixpf:
            for ixp in ixpf.readlines():
                ixpno, ipprefix = ixp.split()
                self.ixps[ixpno].add(ipprefix)

    def name(self, ixpno):
        """Return name for ixpno if known as IXP."""
        if ixpno in self.ixps:
            return self.ixps[ixpno].name
        return None

    def lookup(self, ipstring):
        """Lookup IXP containing ipstring in dotted notation."""
        for (_, ixp) in self.ixps.iteritems():
            for prefix in ixp.prefixes:
                if prefix.contains(ipstring):
                    return ixp
        return None

    def analyze_trace(self, trace, remove_common=False):
        """Analyze Scapy trace for crossed countries and IXPs.

        If remove_common is True, initial common hops are ignored.  You may
        want to call this with its default value False first.  If the first
        hops appear to be safe, a second invocation with True may lead to
        shorter routes with less irrelevant output.
        """
        ips_per_ixp = dict()
        safe_targets = set()
        ipinfos = self.iplookup.lookup(scapy_ipstrings(trace, remove_common))
        for (ipstring, route) in trace.iteritems():
            ipaddr = ipinfos[ipstring]
            hopnos = route.keys()
            hopnos.sort()
            for hopno in hopnos:
                hopipstring = route[hopno][0]
                hopipaddr = ipinfos[hopipstring]
                ipaddr.append_hop(hopipaddr)
                ixp = self.lookup(hopipstring)
                if ixp is not None:
                    ipaddr.add_ixp(ixp)
                    hopipaddr.add_ixp(ixp)
                    msg = "{0}: IXP {1} at hop {2} of {3}".format(
                        ipstring, ixp, hopno, route)
                    logging.info(msg)
                    if not ixp.ixpno in ips_per_ixp:
                        ips_per_ixp[ixp.ixpno] = set()
                    ips_per_ixp[ixp.ixpno].add(ipaddr)
            if not ipaddr.isforeign and not ipaddr.passes_ixp():
                safe_targets.add(ipstring)
        return (ipinfos, safe_targets, ips_per_ixp)


class TorRouter(object):
    """Represent single Tor router."""

    naming = {
        "Router Name": "name",
        "Country Code": "cc",
        "IP Address": "ipstring",
        "Bandwidth (KB/s)": "bandwidth",
        "Uptime (Days)": "uptime"}

    def __init__(self, kwdict):
        for key, value in kwdict.iteritems():
            if key.startswith("Flag -"):
                flag = key.split("-")[1]
                flag = flag.lower().replace(" ", "")
                value = (value == "1")
                setattr(self, "is" + flag, value)
            elif key in TorRouter.naming:
                setattr(self, TorRouter.naming[key], value)
            else:
                setattr(self, key, value)

    def __str__(self):
        return str([self.name, self.ipstring, self.cc])


class TorDB(object):
    """Tor router information imported from CSV."""

    def __init__(self, filename="Tor_query_EXPORT.csv", predicate=None):
        reader = csv.DictReader(open(filename), delimiter=",")
        self.routers = dict()
        total = 0
        for csvdict in reader:
            total += 1
            router = TorRouter(csvdict)
            if predicate is not None and not predicate(router):
                logging.debug("Predicate failed: %s", csvdict)
                continue
            if router.ipstring in self.routers:
                logging.warning(
                    "Overwriting router with same IP address!? %s with %s",
                    self.routers[router.ipstring], router)
            self.routers[router.ipstring] = router
        logging.warning(
            "Parsed %s routers.  Using %s.", total, len(self.routers))

    def lookup(self, ipstring):
        """Return dictionary describing the router with ipstring or None."""
        if ipstring in self.routers:
            return self.routers[ipstring]
        return None

    def get_router_ipstrings(self):
        """Return IP addresses of routers."""
        return self.routers.keys()
