#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright 2013 Jens Lechtenbörger
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Sample usage of ixp.py.

See README.txt.
"""
import ixp
import ixp_config
import ixp_scapy


def print_header(header, objects):
    """Print header and number of objects; return that number."""
    number = len(objects)
    if number == 0:
        print header, "None"
    else:
        print header, number
    return number


def print_routes(header, routers, ipinfos):
    """Print header, then routes to routers based on ipinfos."""
    if print_header(header, routers) > 0:
        routers.sort(key=lambda router: router.name)
        for router in routers:
            print ipinfos[router.ipstring].routestr()


def print_routers(header, routers):
    """Print header, then routers."""
    if print_header(header, routers) > 0:
        routers.sort(key=lambda router: router.name)
        print ",".join(map(lambda router: router.name, routers))


def main():
    ixp_config.init()
    ccode = ixp_config.CONFIG["country"]
    remove_common = ixp_config.CONFIG["remove_common"]
    show_safe = ixp_config.CONFIG["show_safe"]
    ixp.IPAddress.country = ccode

    print "Loading Tor data."
    tordb = ixp.TorDB(
        filename=ixp_config.CONFIG["csvfile"],
        predicate=ixp_config.check_router)
    print "Loaded {0} routers.".format(len(tordb.routers))

    print "\nLoading traceroute data."
    trace = ixp_scapy.get_trace(ipstrings=tordb.routers.keys())
    print "Loaded {0} traces.".format(len(trace))

    print "\nSearching for IXPs."
    options = {
        "ixpfile": ixp_config.dataname("ixp-list"),
        "ixpprefixes": ixp_config.dataname("ixp-prefixes"),
        "pygeoip": ixp_config.dataname("geoip"),
        "cymru": None}
    if ixp_config.CONFIG["torpass"] is not None:
        options["torctl"] = ixp_config.CONFIG["torpass"]
    ixpsearch = ixp.IXPSearch(**options)
    result = ixpsearch.analyze_trace(trace, remove_common=remove_common)
    (ipinfos, safe_targets, ips_per_ixp) = result

    routers_per_as = dict()
    safe_routers = list()
    boomerangs = list()
    otherforeigners = list()
    natives = list()
    for ipstring in tordb.routers:
        ipaddr = ipinfos[ipstring]
        router = tordb.routers[ipstring]
        torname = router.name
        ipaddr.torname = torname
        if ipstring in safe_targets:
            safe_routers.append(router)
            asno = ipaddr.asno
            if asno not in routers_per_as:
                routers_per_as[asno] = set()
            routers_per_as[asno].add(torname)
        if ipaddr.isboomeranged:
            boomerangs.append(router)
        elif ipaddr.isforeign:
            otherforeigners.append(router)
        else:
            natives.append(router)

    ixp_routed = 0
    print """
{0} IXPs found.
Routes to targets per IXP:""".format(len(ips_per_ixp))
    for (ixpno, ipaddrs) in ips_per_ixp.iteritems():
        ixp_routed += len(ipaddrs)
        print "\nIXP {0} ({1} targets):".format(ixpno, len(ipaddrs))
        for ipaddr in ipaddrs:
            print ipaddr.routestr()
    print "(Total: {0})".format(ixp_routed)

    print_routes(
        "\nBoomerang routes (returning from foreign hops):",
        boomerangs, ipinfos)

    print_routes(
        "\nOther routes with foreign hops or targets:",
        otherforeigners, ipinfos)

    if show_safe:
        print_routes("\nSafe routes:", safe_routers, ipinfos)

    print_routers("\nBoomeranged routers:", boomerangs)

    print_routers(
        "\nOther routers with foreign hops or targets:", otherforeigners)

    print_routers("\nNon-foreigners (ignoring IXP data):", natives)

    if len(safe_targets) > 0:
        print """
{0} routers (out of {1}) in {2}, neither passing IXPs nor foreign hops,
grouped by {3} ASes:""".format(
            len(safe_targets), len(tordb.routers), ccode, len(routers_per_as))
        for asno in routers_per_as:
            print "AS {0}: {1}".format(asno, routers_per_as[asno])
        print_routers("\nNames for those routers:", safe_routers)
    else:
        print "\nNo routers found that neither pass IXPs nor foreign hops."


if __name__ == '__main__':
    main()
