#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright 2013 Jens Lechtenbörger
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import getopt
import logging
import os
import sys

CONFIG = {
    "country": "DE",
    "datadir": "data",
    "guards": True,
    "named": True,
    "torpass": None,
    "loglevel": logging.WARNING,
    "geoip": "GeoIP.dat",
    "ixp-list": "ixp-list.txt",
    "ixp-prefixes": "ixp-prefixes.txt",
    "csvfile": "Tor_query_EXPORT.csv",
    "tracefile": "ixp_scapy_trace.py",
    "remove_common": False,
    "show_safe": False}


def dataname(what):
    """Return filename for what.

    If the path for what contains a separator it is returned unchanged.
    Otherwise, the result will be located under a data directory.
    """
    if CONFIG[what].find(os.sep) > -1:
        return CONFIG[what]
    return os.path.join(CONFIG["datadir"], CONFIG[what])


def check_router(router):
    """Check whether router satisfies configured properties."""
    result = not router.isbadexit
    if CONFIG["guards"]:
        result &= router.isguard
    else:
        result &= router.isexit
    if CONFIG["named"]:
        result &= router.isnamed
    result &= (router.cc == CONFIG["country"])
    return result


def usage():
    """Print usage information and exit."""
    print """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Usage: {0} [-c country -l loglevel -e -f trace -r -s -t torpw -u] csvfile
Note that some options are ignored by ixp_scapy.py.
    -l loglevel: default {2}; lower values produce more output
    -c: restrict analysis to given countrycode (default: {1})
    -e: restrict analysis to Tor exits instead of Tor guards
    -f trace: use this file for Scapy traceroutes (default: {2})
    -r: remove hops that are common to all traceroutes (dangerous)
    -s: also show routes to safe routers
    -t torpw: tor control password
    -u: allow unnamed routers among result
    csvfile: CSV file with Tor routers
    """.format(
        sys.argv[0], CONFIG["country"], CONFIG["loglevel"],
        CONFIG["tracefile"])
    sys.exit(0)


def init():
    """Update CONFIG based on command line arguments."""
    try:
        optlist, args = getopt.getopt(
            sys.argv[1:], "c:ef:rst:uhl:", [
                "country_code=", "exit", "file_trace=",
                "remove_common", "show_safe",
                "torpassword=", "unnamed", "help", "log_level="])
    except getopt.GetoptError:
        print "Unknown option."
        usage()
    for opt, arg in optlist:
        if opt in ("-h", "--help"):
            usage()
        elif opt in ("-c", "--country_code"):
            CONFIG["country"] = arg
        elif opt in ("-e", "--exit"):
            CONFIG["guards"] = False
        elif opt in ("-f", "--file_trace"):
            CONFIG["tracefile"] = arg
        elif opt in ("-r", "--remove_common"):
            CONFIG["remove_common"] = True
        elif opt in ("-s", "--show_safe"):
            CONFIG["show_safe"] = True
        elif opt in ("-t", "--torpassword"):
            CONFIG["torpass"] = arg
        elif opt in ("-u", "--unnamed"):
            CONFIG["named"] = False
        elif opt in ("-l", "--log_level"):
            CONFIG["loglevel"] = int(arg)
    if len(args) > 0:
        CONFIG["csvfile"] = args[0]
    else:
        usage()
    log = logging.getLogger()
    log.setLevel(CONFIG["loglevel"])
