#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright 2013 Jens Lechtenbörger
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Generate, save, and read traceroute to and from file.

Scapy's traceroute needs root privileges for raw socket access.
Therefore, this code is singled out.
"""
import ast
import logging
import os
import sys

from scapy.all import traceroute

import ixp
import ixp_config


def store_trace(ipstrings, filename=None):
    """Store traceroute for ipstrings in filename.

    If filename is None, obtain it from ixp_config.
    """
    if filename is None:
        filename = ixp_config.dataname("tracefile")
    if os.path.isfile(filename):
        result = get_trace(filename=filename)
        wanted = set(ipstrings)
        known = set(result.keys())
        unknown = list(wanted - known)
    else:
        result = dict()
        unknown = ipstrings
    if len(unknown) > 0:
        logging.info("Traceroute for %s targets.", len(unknown))
        answer = traceroute(unknown)[0]
        result.update(answer.get_trace())
        with open(filename, "w") as trace:
            trace.write(str(result))
    else:
        msg = """Traces to all {0} targets available.  Not calling scapy.
You must rename the tracefile {1} if you want to obtain
fresh traceroute data.""".format(len(ipstrings), filename)
        logging.warning(msg)


def get_trace(filename=None, ipstrings=None):
    """Read and return traceroute result from filename.

    If filename is None, obtain it from ixp_config.
    If ipstrings is not None, only return traceroute data for those targets.
    """
    if filename is None:
        filename = ixp_config.dataname("tracefile")
    with open(filename) as trace:
        result = ast.literal_eval(trace.read())
    if ipstrings is None:
        return result
    for ipstring in result.keys():
        if ipstring not in ipstrings:
            del result[ipstring]
    for ipstring in ipstrings:
        if ipstring not in result:
            msg = """Traceroute for {0} unknown.
(Did you execute ixp_scapy.py as root with your current options to generate
necessary traceroute data?)""".format(ipstring)
            logging.error(msg)
            sys.exit(1)
    dedup(result)
    return result


def dedup(trace):
    """Remove duplicate hops from trace."""
    for (_, route) in trace.iteritems():
        hops = route.keys()
        hops.sort()
        nohops = len(hops)
        current = nohops - 1
        target = route[hops[current]][0]
        while current > 0 and target == route[hops[current - 1]][0]:
            logging.debug("Deleting hop %s in %s.", hops[current], route)
            del route[hops[current]]
            current -= 1


def main():
    """Perform Scapy traceroute and store result."""
    ixp_config.init()
    tordb = ixp.TorDB(
        filename=ixp_config.CONFIG["csvfile"],
        predicate=ixp_config.check_router)
    store_trace(tordb.get_router_ipstrings())


if __name__ == '__main__':
    main()
